% Social Golfer problem
% http://4c.ucc.ie/~tw/csplib//prob/prob010/index.html
% Copyright (C) 2003 INRIA Sylvain Soliman

% file soc_golf2.pl
% May 07 2003: Version 0.0.3
% sbds version

golf(G,S,W,R) :-     % G groups of S golfers play for W weeks
	M is G*S,
	length(R,W),      % R, the result, is a list of weeks, each being a list...
	%  R = [Week1|WL],
	%  instantiate_week1(Week1,G,S),
	init_weeks(M,G,S,R,Vars), % ... of group numbers attributed to each player
	not_twice_rec(R),
	nth(1,R,FirstWeek),	
	set_first_week(1,G,S,FirstWeek),% breaks week symmetries
	(W>1 ->	nth(2,R,SecWeek),
		set_first_group(SecWeek,S,0);
		true),
%	sym_between_group(G,S,W,LSymWeek,1),
	sym_weeks(G,S,W,LSymWeek),
	sym_group(G,S,1,W,LSymGroup),!,
	flatten([LSymWeek,LSymGroup],LSym),
	write(LSym),nl,
	
	randomize,
	fd_labeling_sbds(Vars,[group(LSym)],[value_method(random)]). 
   
% Dynamic Symmetries

% Symmetries between weeks
sym_weeks(G,S,W,SymWeek):-
	N is G*S,
	Fin is N*(W-1)+1,
	Np1 is N+1,
	enr_weeks_var_1(1,Np1,N,Fin,SymWeek).
	
enr_weeks_var_1(Fin,_,_,Fin,[]):-!.
enr_weeks_var_1(I,J,N,Fin,SymWeek):-
	enr_weeks_var_2(I,J,N,Fin,SymWeek1),

	IpN is I+N,
	JpN is IpN+N,
	enr_weeks_var_1(IpN,JpN,N,Fin,SymWeek2),
	append(SymWeek1,SymWeek2,SymWeek).

enr_weeks_var_2(_,J,_,Fin,[]):-J>Fin,!.
enr_weeks_var_2(I,J,N,Fin,[sym(red,Sym,[])|SymWeek]):-
	IpN is I+N,
	pred_week(I,J,IpN,Sym),
	JpN is J+N,
	enr_weeks_var_2(I,JpN,N,Fin,SymWeek).

pred_week(Fin,_,Fin,[]):-!.
pred_week(I,J,Fin,[[I,J]|L]):-
	Ip1 is I+1,
	Jp1 is J+1,
	pred_week(Ip1,Jp1,Fin,L).


% Symmetries between groups 1

sym_group(_,_,I,W,[]):-I>W,!.
sym_group(G,S,I,W,LSymGroup):-
	N is G*S,
	Deb is (I-1)*N+1,
	Fin is Deb+N-1,
	liste_num(Week,Deb,Fin),
	sym_group_1(Week,1,G,LSymGroup1),
	Ip1 is I+1,
	sym_group(G,S,Ip1,W,LSymGroup2),
	append(LSymGroup1,LSymGroup2,LSymGroup).

sym_group_1(_,G,G,[]):-!.
sym_group_1(Week,I,G,LSymGroup):-
	J is I+1,
	sym_group_11(Week,I,J,G,LSymGroup1),
	Ip1 is I+1,
	sym_group_1(Week,Ip1,G,LSymGroup2),
	append(LSymGroup1,LSymGroup2,LSymGroup).

sym_group_11(_,_,J,G,_):-J>G,!.
sym_group_11(Week,I,J,G,[sym(val_local,(I,J),Week)|LSym]):-
	Jp1 is J+1,
	sym_group_11(Week,I,Jp1,G,LSym).

% Symmetries between groups 2
sym_between_group(G,S,W,LSymWeek,InitPoint):-
	N is G*S,
	NbVar is G*S*W,
	NbVal is G+1,
	group_inside_week(InitPoint,N,LSymWeek,NbVar,NbVal,G,S).
	

group_inside_week(InitPoint,_,[],NbVar,NbVal,_,_):-
	LgPoint is NbVar*NbVal,InitPoint>LgPoint,!.
group_inside_week(InitPoint,FinalPoint,SymWeek,NbVar,NbVal,G,S):-
	LgPoint is NbVar*NbVal,
	
	one_sym_group_G1(InitPoint,FinalPoint,SymWeek1,LgPoint,NbVal,1,G,S),
	
	InitPointNxt is InitPoint+G*S*NbVal,
	group_inside_week(InitPointNxt,FinalPoint,SymWeek2,NbVar,NbVal,G,S),
	flatten([SymWeek1,SymWeek2],SymWeek).

one_sym_group_G1(_,_,[],_,_,G,G,_):-!.
one_sym_group_G1(InitPoint,FinalPoint,SymWeek,LgPoint,NbVal,I,G,S):-
	J is I+1,
	one_sym_group_G1_G2(InitPoint,SymWeek1,LgPoint,NbVal,I,J,G,S),
	one_sym_group_G1(InitPoint,FinalPoint,SymWeek2,LgPoint,NbVal,J,G,S),
	flatten([SymWeek1,SymWeek2],SymWeek).

one_sym_group_G1_G2(_,[],_,_,_,J,G,_):-J>G,!.
one_sym_group_G1_G2(InitPoint,[sym(point,LPoint)|LSym],LgPoint,NbVal,I,J,G,S):-
	
	length(LPoint,LgPoint),
	(InitPoint>1 -> P1 is InitPoint-1,liste_num(LFirst,1,P1);LFirst=[]),
	LgPermute is G*S*NbVal,
	length(LPermute,LgPermute),
	
	enr_LVal(LPermute,InitPoint,I,J,NbVal),
	P2 is InitPoint+LgPermute,
	(P2=<LgPoint ->liste_num(LLast,P2,LgPoint);LLast=[]),
	flatten([LFirst,LPermute,LLast],LPoint),
	Jp1 is J+1,
	one_sym_group_G1_G2(InitPoint,LSym,LgPoint,NbVal,I,Jp1,G,S).
	

% Static Symmetries
set_first([_]).
set_first([X|[Y|R]]):-
	X#=<Y,
	set_first([Y|R]).

set_first_week(_,_,_,[]):-!.
set_first_week(I,G,S,Week):-
	length(Group,S),
	init_group(Group,I),
	Ip1 is I+1,
	append(Group,RWeek,Week),
	set_first_week(Ip1,G,S,RWeek).

init_group([],_).
init_group([X|R],I):-
	X#=I,
	init_group(R,I).

set_first_group(_,S,S):-!.
set_first_group([X|R],S,I) :-
	X#=1,
	length(OneGroup,S),
	append(OneGroup,RGroup,[X|R]),
	Ip1 is I+1,
	set_first_group(RGroup,S,Ip1).
	
% Initialization
init_weeks(_,_,_,[],[]).
init_weeks(M,G,S,[GL|WL],V) :-
	length(GL,M),           % M players in the list
	fd_domain(GL,1,G),      % each player gets a group number
	fd_exactly_rec(GL,G,S), % exactly S golfers in each group
	init_weeks(M,G,S,WL,V1),
	append(GL,V1,V).

fd_exactly_rec(_,0,_).
fd_exactly_rec(L,G,S) :-
   G > 0,
   GG is G-1,
   fd_exactly(S,L,G),      % exactly S golfers in group G
   fd_exactly_rec(L,GG,S).

not_twice_rec([]).
not_twice_rec([W|WL]) :-
   not_twice(W,WL),
   not_twice_rec(WL).

not_twice(_,[]).
not_twice(W,[H|T]) :-
   no_common_rec(W,H),
   not_twice(W,T).

no_common_rec([],[]).
no_common_rec([H1|T1],[H2|T2]) :-
   no_common(H1,T1,H2,T2),
   no_common_rec(T1,T2).

no_common(_,[],_,[]).
no_common(G1,[H1|T1],G2,[H2|T2]) :-
   (H1 #= G1) #==> (G2 #\= H2),
   (H2 #= G2) #==> (G1 #\= H1),
   no_common(G1,T1,G2,T2).
   
